/*
 * This file was generated automatically by ExtUtils::ParseXS version 3.40 from the
 * contents of CHM.xs. Do not edit this file, edit CHM.xs instead.
 *
 *    ANY CHANGES MADE HERE WILL BE LOST!
 *
 */

#line 1 "CHM.xs"
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include "ppport.h"

#include <string.h>

/* Including the chm header */
#include <chm_lib.h>


#define CHM_MAX_BUF 65536

static char * my_strndup(const char *, size_t );

/*
 * This struct will hold some necessary data for the extension and is
 * the opaque C-object returned by new().
 *
 */

typedef struct 
{

  struct chmFile* chmfile;
  char*           filename;

} ChmObj;


/*
 * ChmObjData holds some infos about a particular member of the chm
 * file. It will be used to build a linked list that contains infos
 * about the content of the chm file. The member of the structure are:
 *
 *     - path is the complete path of the member;
 *     - title is the title of the member if it is an html file;
 *     - size is the file-size of the member;
 *     - next is a pointer to the next structure .
 *
 */

struct ChmObjData
{

  char *path;
  char *title;
  size_t size;
  struct ChmObjData *next;

};


/*
 * We must(?) make this variable global, so it could be used in
 * _chm_enumerate_callback().
 *
 */

struct ChmObjData *data = NULL;



/*
 * This function creates a new node for the linked list containing the
 * filelist of the chm file.
 *
 */

static struct ChmObjData *chm_data_add(char *path, char *title, size_t size)
{

  struct ChmObjData *tmp;

  tmp = calloc(1, sizeof(struct ChmObjData));
  if ( NULL == tmp )
    croak("Out of memory\n");

  tmp->path  = my_strndup(path, strlen(path));
  tmp->title = title;
  tmp->size  = size;
  tmp->next  = NULL;
    
  return tmp;

}


/*
 * Init the ChmObject.
 *
 */

static ChmObj *perl_chm_init(const char* filename)
{

  ChmObj *self = NULL;

  self = Newz(1, self, 1, ChmObj);
    
  if ( NULL == self )
    croak("Out of memory");
  
  self->filename = calloc(strlen(filename) + (size_t)1, sizeof(char));
  
  if ( !self->filename )
    croak("Out of memory");
  else
    strncpy(self->filename, filename, strlen(filename));

  self->chmfile = chm_open(self->filename);

  if ( NULL == self->chmfile )
    croak("Can't open file %s\n", self->filename);

  return self;

}



/*
 * Get rid of faulty strndup implementations...
 *
 */

static char * my_strndup(const char *src, size_t len)
{

  char *ret = NULL;

  ret = calloc(len + (size_t)1, sizeof(char));

  if ( !ret )
    croak("Out of memory\n");

  strncpy(ret, src, len);
  *(ret + (int)len) = '\0';

  return ret;

}


/*
 * Given a valid "path", perl_get_object() fetch its content from
 * chmfile, setting len to the length of the string that contains it
 * (len is a value-result argument).
 *
 */

static unsigned char *perl_get_object(struct chmFile *chmfile, const char *path, size_t *len)
{

  struct chmUnitInfo ui;
  unsigned char chm_buffer[CHM_MAX_BUF], *retbuf;
  int offset, swath;

  if ( CHM_RESOLVE_SUCCESS != chm_resolve_object(chmfile, path, &ui) )
    croak("Can't resolve given path\n");

  Newz(1, retbuf, (size_t)ui.length, unsigned char);

  if ( NULL == retbuf )
       croak("Out of memory\n");

  swath = CHM_MAX_BUF;
  offset = 0;
  *len = (size_t)ui.length;  

  while ( offset < ui.length ) /* Let's read the file at swath bytes per time, */
    {                          /* incrementing offset with the data we've read */

      if ( (ui.length - offset) < CHM_MAX_BUF )
	swath = ui.length - offset;
      else
	swath = CHM_MAX_BUF;

      swath = (int)chm_retrieve_object(chmfile, &ui, chm_buffer, offset, swath);
      memcpy((retbuf + offset), chm_buffer, (size_t)swath);
      offset += swath;

    }
  
  return retbuf;

}


static int file_is_normal(const char *filename)
{

  /* 
   * For the moment, we are intrested only on normal files, such as
   * html, css, images, etc... So we limit to look to these.
   *
   */
  
  if ( '/' == *filename ) /* SPECIAL or NORMAL file */
    {
      if ( ('#' == *(filename + 1)) || ('$' == *(filename + 1)) ) /* SPECIAL file */
	return 0;
      else /* NORMAL file */
	return 1;
    }
  else /* META file */
    return 0;

}


/*
 * Check that the file at "path" is an html one, taking this decision
 * depending on the extension of the file.
 *
 */

static int file_is_html(const char *path)
{

  char *tmp;

  if ( (int)strlen(path) < 4 ) /* Path must be at least 5 char long. */
    return 0;

  if ( NULL != (tmp = strrchr(path, '.')) )
    if ( 0 == strncasecmp(++tmp, "htm", (size_t)3) || 
	 0 == strncasecmp(tmp, "html", (size_t)4) )
	return 1;

  return 0;

}


/*
 * Traverse the content of the string "s", which contains html code,
 * to find the text between <title> and </title>, and return it.
 *
 */

static char *find_title(char *s)
{

  char *tmp = s;
  size_t len;

  while ( tmp++ )
    {
      tmp = strchr(tmp, '<');
      
      if ( 0 == strncasecmp(tmp, "<title>", (size_t)7) )
	{
	  tmp += 7;
	  len = (size_t) (strchr(tmp, '<') - tmp);
	  return my_strndup(tmp, len);
	}
      else
	continue;
    }

  return NULL;
	  
}      


/*
 * Glue-function for getting the title of an html file and returning
 * it.
 *
 */

static char *get_html_title(struct chmFile *chmfile, const char *filename)
{

  char *content, *title;
  size_t len;
  
  content = perl_get_object(chmfile, filename, &len);
  title = find_title(content);

  return title;

}


/*
 * Callback function to be passed to chm_enumerate(), in order to
 * build a linked list of ChmObjData structures that contains some
 * informations about the "normal" members (html, css, xml and image
 * files) of the chm file.
 *
 */

static int _chm_enumerate_callback(struct chmFile *h, struct chmUnitInfo *ui, void *context)
{
  
  char *title;
  struct ChmObjData *tmp;

  if ( file_is_normal(ui->path) )
    {
      title = ((file_is_html(ui->path)) ? get_html_title(h, ui->path) : NULL);
      tmp = chm_data_add(ui->path, title, ui->length);

      data->next = tmp;
      data = data->next;    
      tmp = NULL;
    }

  return CHM_ENUMERATOR_CONTINUE;

}
  

#line 327 "CHM.c"
#ifndef PERL_UNUSED_VAR
#  define PERL_UNUSED_VAR(var) if (0) var = var
#endif

#ifndef dVAR
#  define dVAR		dNOOP
#endif


/* This stuff is not part of the API! You have been warned. */
#ifndef PERL_VERSION_DECIMAL
#  define PERL_VERSION_DECIMAL(r,v,s) (r*1000000 + v*1000 + s)
#endif
#ifndef PERL_DECIMAL_VERSION
#  define PERL_DECIMAL_VERSION \
	  PERL_VERSION_DECIMAL(PERL_REVISION,PERL_VERSION,PERL_SUBVERSION)
#endif
#ifndef PERL_VERSION_GE
#  define PERL_VERSION_GE(r,v,s) \
	  (PERL_DECIMAL_VERSION >= PERL_VERSION_DECIMAL(r,v,s))
#endif
#ifndef PERL_VERSION_LE
#  define PERL_VERSION_LE(r,v,s) \
	  (PERL_DECIMAL_VERSION <= PERL_VERSION_DECIMAL(r,v,s))
#endif

/* XS_INTERNAL is the explicit static-linkage variant of the default
 * XS macro.
 *
 * XS_EXTERNAL is the same as XS_INTERNAL except it does not include
 * "STATIC", ie. it exports XSUB symbols. You probably don't want that
 * for anything but the BOOT XSUB.
 *
 * See XSUB.h in core!
 */


/* TODO: This might be compatible further back than 5.10.0. */
#if PERL_VERSION_GE(5, 10, 0) && PERL_VERSION_LE(5, 15, 1)
#  undef XS_EXTERNAL
#  undef XS_INTERNAL
#  if defined(__CYGWIN__) && defined(USE_DYNAMIC_LOADING)
#    define XS_EXTERNAL(name) __declspec(dllexport) XSPROTO(name)
#    define XS_INTERNAL(name) STATIC XSPROTO(name)
#  endif
#  if defined(__SYMBIAN32__)
#    define XS_EXTERNAL(name) EXPORT_C XSPROTO(name)
#    define XS_INTERNAL(name) EXPORT_C STATIC XSPROTO(name)
#  endif
#  ifndef XS_EXTERNAL
#    if defined(HASATTRIBUTE_UNUSED) && !defined(__cplusplus)
#      define XS_EXTERNAL(name) void name(pTHX_ CV* cv __attribute__unused__)
#      define XS_INTERNAL(name) STATIC void name(pTHX_ CV* cv __attribute__unused__)
#    else
#      ifdef __cplusplus
#        define XS_EXTERNAL(name) extern "C" XSPROTO(name)
#        define XS_INTERNAL(name) static XSPROTO(name)
#      else
#        define XS_EXTERNAL(name) XSPROTO(name)
#        define XS_INTERNAL(name) STATIC XSPROTO(name)
#      endif
#    endif
#  endif
#endif

/* perl >= 5.10.0 && perl <= 5.15.1 */


/* The XS_EXTERNAL macro is used for functions that must not be static
 * like the boot XSUB of a module. If perl didn't have an XS_EXTERNAL
 * macro defined, the best we can do is assume XS is the same.
 * Dito for XS_INTERNAL.
 */
#ifndef XS_EXTERNAL
#  define XS_EXTERNAL(name) XS(name)
#endif
#ifndef XS_INTERNAL
#  define XS_INTERNAL(name) XS(name)
#endif

/* Now, finally, after all this mess, we want an ExtUtils::ParseXS
 * internal macro that we're free to redefine for varying linkage due
 * to the EXPORT_XSUB_SYMBOLS XS keyword. This is internal, use
 * XS_EXTERNAL(name) or XS_INTERNAL(name) in your code if you need to!
 */

#undef XS_EUPXS
#if defined(PERL_EUPXS_ALWAYS_EXPORT)
#  define XS_EUPXS(name) XS_EXTERNAL(name)
#else
   /* default to internal */
#  define XS_EUPXS(name) XS_INTERNAL(name)
#endif

#ifndef PERL_ARGS_ASSERT_CROAK_XS_USAGE
#define PERL_ARGS_ASSERT_CROAK_XS_USAGE assert(cv); assert(params)

/* prototype to pass -Wmissing-prototypes */
STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params);

STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params)
{
    const GV *const gv = CvGV(cv);

    PERL_ARGS_ASSERT_CROAK_XS_USAGE;

    if (gv) {
        const char *const gvname = GvNAME(gv);
        const HV *const stash = GvSTASH(gv);
        const char *const hvname = stash ? HvNAME(stash) : NULL;

        if (hvname)
	    Perl_croak_nocontext("Usage: %s::%s(%s)", hvname, gvname, params);
        else
	    Perl_croak_nocontext("Usage: %s(%s)", gvname, params);
    } else {
        /* Pants. I don't think that it should be possible to get here. */
	Perl_croak_nocontext("Usage: CODE(0x%" UVxf ")(%s)", PTR2UV(cv), params);
    }
}
#undef  PERL_ARGS_ASSERT_CROAK_XS_USAGE

#define croak_xs_usage        S_croak_xs_usage

#endif

/* NOTE: the prototype of newXSproto() is different in versions of perls,
 * so we define a portable version of newXSproto()
 */
#ifdef newXS_flags
#define newXSproto_portable(name, c_impl, file, proto) newXS_flags(name, c_impl, file, proto, 0)
#else
#define newXSproto_portable(name, c_impl, file, proto) (PL_Sv=(SV*)newXS(name, c_impl, file), sv_setpv(PL_Sv, proto), (CV*)PL_Sv)
#endif /* !defined(newXS_flags) */

#if PERL_VERSION_LE(5, 21, 5)
#  define newXS_deffile(a,b) Perl_newXS(aTHX_ a,b,file)
#else
#  define newXS_deffile(a,b) Perl_newXS_deffile(aTHX_ a,b)
#endif

#line 471 "CHM.c"

XS_EUPXS(XS_Text__CHM_new); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Text__CHM_new)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "CLASS, file");
    {
	char *	CLASS = (char *)SvPV_nolen(ST(0))
;
	char *	file = (char *)SvPV_nolen(ST(1))
;
	ChmObj *	RETVAL;
#line 326 "CHM.xs"
     RETVAL = perl_chm_init(file);

     if ( NULL == RETVAL )
       XSRETURN_UNDEF;
#line 490 "CHM.c"
	{
	    SV * RETVALSV;
	    RETVALSV = sv_newmortal();
	    sv_setref_pv( RETVALSV, CLASS, (void*)RETVAL );
	    ST(0) = RETVALSV;
	}
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Text__CHM_DESTROY); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Text__CHM_DESTROY)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	ChmObj *	self;

        if( sv_isobject(ST(0)) && (SvTYPE(SvRV(ST(0))) == SVt_PVMG) )
                self = (ChmObj *)SvIV((SV*)SvRV( ST(0) ));
        else{
                warn( "Text::CHM::DESTROY() -- self is not a blessed SV reference" );
                XSRETURN_UNDEF;
        }
;
#line 338 "CHM.xs"
     chm_close(self->chmfile);
     safefree(self);
#line 521 "CHM.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Text__CHM_filename); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Text__CHM_filename)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	ChmObj *	self;
	char *	RETVAL;
	dXSTARG;

        if( sv_isobject(ST(0)) && (SvTYPE(SvRV(ST(0))) == SVt_PVMG) )
                self = (ChmObj *)SvIV((SV*)SvRV( ST(0) ));
        else{
                warn( "Text::CHM::filename() -- self is not a blessed SV reference" );
                XSRETURN_UNDEF;
        }
;
#line 346 "CHM.xs"
     RETVAL = self->filename;
#line 547 "CHM.c"
	sv_setpv(TARG, RETVAL); XSprePUSH; PUSHTARG;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Text__CHM_close); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Text__CHM_close)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	ChmObj *	self;

        if( sv_isobject(ST(0)) && (SvTYPE(SvRV(ST(0))) == SVt_PVMG) )
                self = (ChmObj *)SvIV((SV*)SvRV( ST(0) ));
        else{
                warn( "Text::CHM::close() -- self is not a blessed SV reference" );
                XSRETURN_UNDEF;
        }
;
#line 355 "CHM.xs"
     chm_close(self->chmfile);
#line 572 "CHM.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Text__CHM_get_object); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Text__CHM_get_object)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "self, objname");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	ChmObj *	self;
	char *	objname = (char *)SvPV_nolen(ST(1))
;
	void *	RETVAL;
	dXSTARG;

        if( sv_isobject(ST(0)) && (SvTYPE(SvRV(ST(0))) == SVt_PVMG) )
                self = (ChmObj *)SvIV((SV*)SvRV( ST(0) ));
        else{
                warn( "Text::CHM::get_object() -- self is not a blessed SV reference" );
                XSRETURN_UNDEF;
        }
;
#line 363 "CHM.xs"
     unsigned char * buf;
     size_t len;
#line 603 "CHM.c"
#line 366 "CHM.xs"
     buf = perl_get_object(self->chmfile, objname, &len);
     XPUSHs(sv_2mortal(newSVpv(buf, len)));
#line 607 "CHM.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Text__CHM_get_filelist); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Text__CHM_get_filelist)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	ChmObj *	self;
	void *	RETVAL;
	dXSTARG;

        if( sv_isobject(ST(0)) && (SvTYPE(SvRV(ST(0))) == SVt_PVMG) )
                self = (ChmObj *)SvIV((SV*)SvRV( ST(0) ));
        else{
                warn( "Text::CHM::get_filelist() -- self is not a blessed SV reference" );
                XSRETURN_UNDEF;
        }
;
#line 374 "CHM.xs"
     struct ChmObjData *contents = NULL;
     HV * hash;
#line 637 "CHM.c"
#line 378 "CHM.xs"
     contents = chm_data_add("start", "start", 0);
     data = contents;

     if ( !chm_enumerate(self->chmfile, CHM_ENUMERATE_ALL, _chm_enumerate_callback, NULL ) )
        croak("Errors getting filelist\n");

     data = contents->next;

     while ( data )
         {
	    hash = newHV();
	    hv_store(hash, "path", 4, newSVpv(data->path, strlen(data->path)), 0);
	    hv_store(hash, "size", 4, newSViv(data->size), 0);
	    if ( data->title )
	      hv_store(hash, "title", 5, newSVpv(data->title, strlen(data->title)), 0);
	    else
	      hv_store(hash, "title", 5, newSV(0), 0);

	    XPUSHs(sv_2mortal(newRV((SV *)hash)));

	    data = data->next;
	 }
#line 661 "CHM.c"
	PUTBACK;
	return;
    }
}

#ifdef __cplusplus
extern "C"
#endif
XS_EXTERNAL(boot_Text__CHM); /* prototype to pass -Wmissing-prototypes */
XS_EXTERNAL(boot_Text__CHM)
{
#if PERL_VERSION_LE(5, 21, 5)
    dVAR; dXSARGS;
#else
    dVAR; dXSBOOTARGSXSAPIVERCHK;
#endif
#if (PERL_REVISION == 5 && PERL_VERSION < 9)
    char* file = __FILE__;
#else
    const char* file = __FILE__;
#endif

    PERL_UNUSED_VAR(file);

    PERL_UNUSED_VAR(cv); /* -W */
    PERL_UNUSED_VAR(items); /* -W */
#if PERL_VERSION_LE(5, 21, 5)
    XS_VERSION_BOOTCHECK;
#  ifdef XS_APIVERSION_BOOTCHECK
    XS_APIVERSION_BOOTCHECK;
#  endif
#endif

        newXS_deffile("Text::CHM::new", XS_Text__CHM_new);
        newXS_deffile("Text::CHM::DESTROY", XS_Text__CHM_DESTROY);
        newXS_deffile("Text::CHM::filename", XS_Text__CHM_filename);
        newXS_deffile("Text::CHM::close", XS_Text__CHM_close);
        newXS_deffile("Text::CHM::get_object", XS_Text__CHM_get_object);
        newXS_deffile("Text::CHM::get_filelist", XS_Text__CHM_get_filelist);
#if PERL_VERSION_LE(5, 21, 5)
#  if PERL_VERSION_GE(5, 9, 0)
    if (PL_unitcheckav)
        call_list(PL_scopestack_ix, PL_unitcheckav);
#  endif
    XSRETURN_YES;
#else
    Perl_xs_boot_epilog(aTHX_ ax);
#endif
}

